
/*****************************************\
*                                         *
* GRFDiff - A program to compare a .grf   *
*           file and the decoded .pcx     *
*           file(s) and produce a file    *
*	    of the sprites that were	  *
*	    modified			  *
*                                         *
*                                         *
* Copyright (C) 2002 by Josef Drexler     *
*                      <jdrexler@uwo.ca>  *
*                                         *
* Permission granted to copy and redist-  *
* ribute under the terms of the GNU GPL.  *
* For more info please read the file      *
* COPYING which should have come with     *
* this file.                              *
*                                         *
\*****************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#include <values.h>
#include <alloc.h>	// remove after testing

#define DOCHECK

#include "myalloc.h"
#include "error.h"
#include "sprites.h"
#include "grfcomm.h"
#include "grfmrg.h"
#include "crc32.h"
#include "path.h"

/*
	Format of a .GRD file:

	Offset	Size	Content
	0	4	Magic code to identify the GRD file
	4	2	GRD file version
	6	2	Number of sprites in this file
	8	1	Length of GRF filename=L (including terminating NULL)
	9	L	GRF filename, without .grf extension or path
	9+L	var.	Sprite data

	Each sprite has this format:

	Offset	Size 	Content
	0	2	Number of the sprite, counting from 0
	2	var.	Sprite content as stored in a GRF file

*/

U32 GRDmagic = 0x67fb49ad;

char *grffile[2] = { NULL, NULL }, *grdfile = NULL;
int selfextr = 0, alwaysyes = 0, onlylist = 0;

char *difflist = NULL;
int difflistfrom = -1, difflistto;

int differences[32][2], diffruns = -1;


char *usagetext=
	"%sUsage:\n"
	"    GRFDiff [options] <Org. GRF-File> <New GRF-File>\n"
	"	Compare the two GRF files and produce a GRD file containing only the\n"
	"	sprites from the second file that are different from the first file.\n"
	"\n"
	"GRFDiff makes a .GRD file with the same basename as the new GRF file.\n"
	"\n"
	"Options:\n"
	"	-h   Show this help\n"
	"	-l <Numbers>\n"
	"	     Save the sprites with these numbers in the .GRD instead of\n"
	"	     finding the modified sprites.  With this option, the first\n"
	"	     GRF file can be omitted.\n"
	"	     Format of the numbers: <from1>[-<to1>][,<from2>[-<to2>]]...\n"
	"		e.g.  1-5,8,20-31 (must be increasing numbers)\n"
	"	-n   Only show a list of modified sprites, don't make a .GRD file\n"
	"	-o <GRD-File>\n"
	"	     Write to this file instead\n"
	"	-x   Make a self-extracting (.EXE) file instead of a .GRD\n"
	"	-y   Answer 'y' to all questions\n"
	"\n"
	"GRFDiff is Copyright (C) 2002 by Josef Drexler <jdrexler@uwo.ca>\n"
	"You may copy and redistribute it under the terms of the GNU General Public\n"
	"License, as stated in the file 'COPYING'.\n";


class infostorer : virtual public spriteinfowriter {
	public:
	infostorer(U8 *infptr, U8 **dataptr, U16 *datasizeptr);

	virtual void addsprite(int x, U8 info[8]);
	virtual void adddata(U16 newsize, U8 *newdata);

	U8 *inf;
	U8 **data;
	U16 *datasize;
};

infostorer::infostorer(U8 *infptr, U8 **dataptr, U16 *datasizeptr)
{
  inf = infptr;
  datasize = datasizeptr;
  data = dataptr;
}

void infostorer::addsprite(int x, U8 info[8])
{
  int i;
  for(i=0; i<8; i++) inf[i] = info[i];
}

void infostorer::adddata(U16 newsize, U8 *newdata)
{
  *datasize = newsize;
  *data = newdata;
}

class grfstore : virtual public spritestorage {
	public:
	grfstore(FILE *f) { grf = f; sdata = NULL; };

	virtual void newsprite();
        virtual void setsize(int sx, int sy);
	virtual void nextpixel(U8 colour);
	virtual void spritedone();

	void getorgdata(U16 *size, U8 **data);

	int differsfrom(grfstore *other);

	FILE *grf;
	int sizex, sizey;
	long spritesize, datasize, spriteofs;
	long fpos_start, fpos_end;
	U8 *sdata;
};

void grfstore::newsprite()
{
  if (sdata)
	free(sdata);
  sdata = NULL;
  fpos_start = ftell(grf);
}

void grfstore::setsize(int sx, int sy)
{
  sizex = sx;
  sizey = sy;
  spritesize = (long) sx * (long) sy;
  spriteofs = 0;

  if (!spritesize)	// verbatim data?
	return;

  sdata = (U8*) myalloc(spritesize);
  if (!sdata) {
	printf("Cannot allocate %ld bytes for sprite\n", spritesize);
	exit(2);
  }
}

void grfstore::nextpixel(U8 colour)
{
  if (spriteofs > spritesize) {
	printf("Sprite has too many pixels!\n");
	exit(2);
  }
  sdata[spriteofs++] = colour;
}

void grfstore::spritedone()
{
  if (spriteofs < spritesize) {
	printf("Sprite has too few pixels!\n");
	exit(2);
  }
  fpos_end = ftell(grf);
}

int grfstore::differsfrom(grfstore *other)
{
  if ( (sizex != other->sizex) || (sizey != other->sizey) )
	return 1;
  return memcmp(sdata, other->sdata, spritesize);
}

void grfstore::getorgdata(U16 *size, U8 **data)
{
  *size = fpos_end - fpos_start;
  long fpos_cur = ftell(grf);

  *data = (U8*) myalloc(*size);
  if (!*data) {
	printf("Cannot allocate %d bytes for data\n", *size);
	exit(2);
  }

  fseek(grf, fpos_start, SEEK_SET);
  fread(*data, 1, *size, grf);
  fseek(grf, fpos_cur, SEEK_SET);
}

char *basename(char *filename)
{
  static char bname[32];
  fnsplit(filename, NULL, NULL, bname, NULL);
  return bname;
}

time_t mtime(FILE *f)
{
  struct stat statbuf;

  fstat(fileno(f), &statbuf);

  return statbuf.st_mtime;
}

char *strmtime(time_t mtime)
{
  static char strtime[128];

  strftime(strtime, sizeof(strtime), "%b %d %Y  %H:%M", localtime(&mtime));

  return strtime;
}

void mkselfextr(FILE *f)
{
  char *block;
  U8 r = 0, e;
  long newr, chunk, grdofs, blank;
  U32 crc;

  crc = crc32(&grfmrg, grfmrgsize);
  if (crc != grfmrgcrc) {
	printf("Warning, self-extracting executable code is damaged.\n"
		"Wanted CRC %lX  Got CRC %lX\n", grfmrgcrc, crc);
	exit(2);
  }

  fwrite(&grfmrg, 1, grfmrgsize, f);

  e = 0;

  while (!r) {
	chunk = 1<<e;

	// round size up to next chunksize
	newr = (long) (grfmrgsize+chunk-1) / chunk;

	if (newr > 255) {
		e++;
	} else {
		r = newr;
	}
  }

  fseek(f, 0x1c, SEEK_SET);

  fwrite("JD", 2, 1, f);
  fwrite(&r, 1, 1, f);
  fwrite(&e, 1, 1, f);

  fseek(f, 0, SEEK_END);
  if (ftell(f) != grfmrgsize) {
	printf("Huh???\n");
	exit(2);
  }

  grdofs = newr * chunk;
  blank = grdofs - grfmrgsize;

  if (blank) {
	block = (char*) myalloc(blank);
	if (!block) {
		printf("Out of memory.\n");
		exit(2);
	}

	memset(block, 0, blank);
	fwrite(block, 1, blank, f);

	free(block);
  }

}

int mkdiff()
{
  FILE *grf[2], *grd;
  long grfsize;
  int res[2], i, j, isdiff, numdiff = 0, numdiffofs, lastpct = -1;
  time_t modtime[2];

  for (i=0; i<2; i++) if (grffile[i]) {
	grf[i] = fopen(grffile[i], "rb");
	if (!grf[i]) {
		fperror(e_openfile, grffile[i]);
		exit(2);
	}
	modtime[i] = mtime(grf[i]);
  } else
	grf[i] = NULL;

  if (grffile[0] && (modtime[0] > modtime[1])) {
	printf("Warning, %s is newer than %s.\n", grffile[0], grffile[1]);
	printf("%s was last modified on %s\n", grffile[0], strmtime(modtime[0]));
	printf("%s was last modified on %s\n", grffile[1], strmtime(modtime[1]));
	printf("But you've told me that %s contains the new sprites.\n", grffile[1]);
	printf("Are you sure you have the right order on the command line [Y/N] ? ");

	if (alwaysyes)
		printf("Y\n");
	else if (tolower(getc(stdin)) != 'y') {
		printf("\nAborted.\n");
		exit(2);
	}
	printf("\nContinuing.\n");
  }

  if (grdfile) {
	grd = fopen(grdfile, "wb");
	if (!grd) {
		fperror(e_openfile, grffile[i]);
		exit(2);
	}
  } else if (!onlylist) {
	doopen(grffile[1], "", selfextr?".EXE":".GRD", "wb", &grdfile, &grd, 0);
  } else {
	grd = NULL;
  }

  if (grd) {
	char *noext = strlwr(basename(grffile[1]));

	if (selfextr)
		mkselfextr(grd);

	fwrite(&GRDmagic, 4, 1, grd);

	i = 1;	// GRD file version
	fwrite(&i, 2, 1, grd);

	numdiffofs = ftell(grd);
	fwrite(&numdiff, 2, 1, grd);

	i = strlen(noext) + 1;
	fwrite(&i, 1, 1, grd);

	fwrite(noext, 1, i, grd);

	printf("Writing %s\n", grdfile);
  } else
	grdfile = NULL;

  fseek(grf[1], 0, SEEK_END);
  grfsize = ftell(grf[1]);
  fseek(grf[1], 0, SEEK_SET);

  U8 inf[2][8], *verbdata[2], *data;
  U16 verbsize[2], size;

  grfstore *grfdata[2];
  infostorer *grfinf[2];

  for (i=0; i<2; i++) if (grf[i]) {
	grfdata[i] = new grfstore(grf[i]);
	grfinf[i] = new infostorer(inf[i], &verbdata[i], &verbsize[i]);
  }

  for (int spriteno = 0; ; spriteno++) {
	int thispct = ftell(grf[1])*100L/grfsize;
	if (thispct != lastpct) {
		lastpct = thispct;
		printf("\rSprite%5d  Done:%3d%%  \r", spriteno, thispct);
	}

	verbdata[0] = verbdata[1] = data = NULL;
	for (i=0; i<2; i++) if (grf[i]) {
		verbsize[i] = 0;
		res[i] = decodesprite(grf[i], grfdata[i], grfinf[i]);
	}

	if (grf[0] && (res[0] != res[1])) {
		printf("The GRF files have a different number of sprites!\n");
		exit(2);
	}

	if (!res[1]) {
		printf("\rSprite%5d  Done:%3d%%  \r", spriteno, 100);
		break;
	}

	if (difflist) {
		while ( (difflistfrom < 0) || (spriteno > difflistto) ) {
			difflistfrom = strtol(difflist, &difflist, 0);
			difflistto = -1;
			switch (difflist[0]) {
				case ',':
					difflist++;
					difflistto = difflistfrom;
					break;
				case 0:
					difflistfrom = difflistto = 32767;
					break;
				case '-':
					difflist++;
					difflistto = strtol(difflist, &difflist, 0);
					if (!difflist[0] || (difflist[0] == ','))
						break;
				default:
					printf("Invalid number format at %s\n", difflist);
					exit(2);
			}
		}
		isdiff = ( (spriteno >= difflistfrom) && (spriteno <= difflistto) );
	} else {
		isdiff = 0;

		// find out if they're different, and remember original data if so
		if (verbsize[0]) {
			if (verbsize[0] != verbsize[1])
				isdiff = 1;
			else
				isdiff = memcmp(verbdata[0], verbdata[1], verbsize[0]);

		} else {
			isdiff = grfdata[0]->differsfrom(grfdata[1]);
		}
	}

	if (isdiff) {
		numdiff++;

		if (!onlylist) {
			// this is a bit redundant for the verbatim data,
			// but the code is a lot clearer this way
			grfdata[1]->getorgdata(&size, &data);

                        // store the modified sprite
			fwrite(&spriteno, 2, 1, grd);
			fwrite(data, 1, size, grd);
		}

		if ( (diffruns < 0) || (spriteno != differences[diffruns][1]+1) ) {
			diffruns++;
			int nummax = sizeof(differences)/sizeof(differences[0]) - 1;
			if (diffruns >= nummax) {
				diffruns = nummax;
				differences[diffruns][0] = -1;
			} else
				differences[diffruns][0] = spriteno;
		}
		differences[diffruns][1] = spriteno;
	}

	if (data) free(data);
	if (verbdata[0]) free(verbdata[0]);
	if (verbdata[1]) free(verbdata[1]);
  }

  if (!onlylist) {
	fseek(grd, numdiffofs, SEEK_SET);
	fwrite(&numdiff, 2, 1, grd);
	fclose(grd);
  }
  if (diffruns < 0) {
	printf("\nNo differences, removing %s", grdfile);
	remove(grdfile);
  } else {
	printf("\nSprites with differences: ");
	for (i=0; i<=diffruns; i++) {
		if (differences[i][0] >= 0) {
			printf("%d", differences[i][0]);
			if (differences[i][0] < differences[i][1])
				printf("-%d", differences[i][1]);
		} else
			printf("... and others");

		if (i < diffruns)
			printf(", ");
	}
        printf(" (%d total)", numdiff);
  }

  for (i=0; i<2; i++)
	if (grf[i]) fclose(grf[i]);

  printf("\nDone!\n");
  return 1;
}       

int main(int argc, char **argv)
{

  printf("GRFDiff - Copyright (C) 2002 by Josef Drexler\n");

  // parse option arguments
  while (1) {
	char opt = getopt(argc, argv, "hl:no:xy");

	if (opt == (char) EOF)
		break;

	switch (opt) {
		case 'l':
			difflist = optarg;
			break;
		case 'n':
			onlylist = 1;
			break;
		case 'o':
			grdfile = optarg;
			break;
		case 'x':
			selfextr = 1;
			break;
		case 'y':
			alwaysyes = 1;
			break;
		default:
			usage();
	}
  }

  // non-option arguments: filename and potentially directory
  if (optind < argc)
	grffile[0] = argv[optind++];
  if (optind < argc)
	grffile[1] = argv[optind++];

  // only one file specified while using -l
  if (difflist && !grffile[1]) {
	grffile[1] = grffile[0];
	grffile[0] = NULL;
  }

  if (!grffile[0] || (strlen(grffile[0]) < 1)
	|| (!difflist && (!grffile[1] || (strlen(grffile[1]) < 1))))
	usage();

  return mkdiff();
}
