
/*****************************************\
*                                         *
* GRFMerge - A program to integrate a     *
*            .GRD file generated by       *
*            GRFDiff into the respective  *
*	     GRF file.			  *
*                                         *
*                                         *
* Copyright (C) 2002 by Josef Drexler     *
*                      <jdrexler@uwo.ca>  *
*                                         *
* Permission granted to copy and redist-  *
* ribute under the terms of the GNU GPL.  *
* For more info please read the file      *
* COPYING which should have come with     *
* this file.                              *
*                                         *
\*****************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "typesize.h"

U32 GRDmagic = 0x67fb49ad;
char *tempfile = "grfmerge.tmp";

int alwaysyes = 0;
char *grdfile = NULL, *grffile = NULL;
long grdofs = 0;
int onlyshow = 0, issfx = 0;

void usage(void)
{
  printf(
	"Usage:\n"
	"    GRFMerge [options] %s[<GRF-File>]\n"
	"	Change sprites in the GRF file to the new ones from the GRD file.\n"
	"	If the GRF file is not specified, GRFMerge will modify the one\n"
	"	which the GRD file was generated from.\n"
	"\n"
	"Options:\n"
	"	-h  Show this help\n"
	"	-l  Only show which sprites the GRD file contains, don't integrate them\n"
	"	-y  Answer 'y' to all questions\n"
	"\n"
	"GRFMerge is Copyright (C) 2002 by Josef Drexler <jdrexler@uwo.ca>\n"
	"It may be freely copied and distributed.\n",
		issfx?"":"<GRD-File> "
  );

  exit(1);
}

int checkisselfextr(char *exe)
{
  FILE *f;
  char c[3];
  U8 r, e;
  U32 magic = 0;

  f = fopen(exe, "rb");
  if (!f) return 0;

  fseek(f, 0x1c, SEEK_SET);
  fread(c, 2, 1, f);
  c[2] = 0;

  if (!strcmp(c, "JD")) {
	fread(&r, 1, 1, f);
	fread(&e, 1, 1, f);

	grdofs = (long) r * (1L << e);
	fseek(f, grdofs, SEEK_SET);

	fread(&magic, 4, 1, f);
  }
  fclose(f);

  return issfx = magic == GRDmagic;
}

void die(char *text, ...)
{
  va_list argptr;

  va_start(argptr, text);
  vprintf(text, argptr);
  va_end(argptr);

  exit(2);
}

void yesno(void)
{
  printf(" [Y/N] ");
  if (alwaysyes)
	printf("Y\n");
  else if (tolower(getc(stdin)) != 'y')
	die("\nAborted.\n");
}

static char *block = NULL;
#define BLOCKSIZE 8192
void copyblock(size_t size, FILE *from, FILE *to)
{
  size_t thisblock;

  while (size > 0) {
	if (size > BLOCKSIZE)
		thisblock = BLOCKSIZE;
	else
		thisblock = size;

	fread(block, 1, thisblock, from);

	if (to) fwrite(block, 1, thisblock, to);

	size -= thisblock;
  }
}

// this function reads a piece of data, and copies it to the other files
void copy(void *data, size_t size, size_t n, FILE *from, FILE *to)
{
  fread(data, size, n, from);

  if (to) fwrite(data, size, n, to);
}

//
// Copy the data of a sprite from one file to another
// this is complicated because for some sprites, the GRF file stores
// the uncompressed length instead of how many bytes are in the file
//
// This code mostly copied from sprites.c, but to keep this program
// small I didn't link in sprites.c and removed the unnecessary steps
// from decodesprite (we're not actually interested in the uncompressed data,
// just the length)
//

void copysprite(FILE *from, FILE *to)
{
  U8 info, ofs;
  S8 code;
  U16 size;
  size_t reallen;

  copy(&size, 2, 1, from, to);
  copy(&info, 1, 1, from, to);

  if (info == 0xff) {	// verbatim data
	copyblock(size, from, to);
	return;
  }

  if (info & 2) {	// size is compressed size
	copyblock(size+7, from, to);	// also copy remaining 7 info bytes
	return;
  }

  // the size given is the uncompressed size, so we need to go through
  // the uncompression routine until we have enough bytes
  copyblock(7, from, to);
  size -= 8;
  while (size > 0) {
	copy(&code, 1, 1, from, to);

	if (code < 0) {
		copy(&ofs, 1, 1, from, to);
		reallen = -(code >> 3);
	} else {
		reallen = code ? code : 128;
		copyblock(reallen, from, to);
	}
	size -= reallen;
  }
}

void skipsprite(FILE *f)
{
  copysprite(f, NULL);
}

int domerge(void)
{
  FILE *grd, *grf, *tmp;
  char grflen, *grfname, *c;
  U16 version, i, j, numsprites, spriteno, curno;
  U32 magic;
  int lastfrom = -2, lastto = lastfrom;

  grd = fopen(grdfile, "rb");
  if (!grd) die("Can't open %s: %s\n", grdfile, strerror(errno));

  fseek(grd, grdofs, SEEK_SET);
  fread(&magic, 4, 1, grd);

  if (magic != GRDmagic) die("This is not a GRD file.\n");

  fread(&version, 2, 1, grd);

  if (version > 1) die("This is a GRD file version %d, I don't know how to handle that.\n", version);

  fread(&numsprites, 2, 1, grd);
  fread(&grflen, 1, 1, grd);

  grfname = malloc(grflen + 4);		// +4 for .bak extension (safety margin)
  if (!grfname) die("Out of memory.\n");

  fread(grfname, 1, grflen, grd);

  if (onlyshow) {
	printf("Generated from: %s.grf\nSprites in file: ", grfname);
	grf = tmp = NULL;
  } else {
	if (grffile) {
		i=strrchr(grffile, '\\') - grffile;
		j=strrchr(grffile+i, '.') - (grffile+i);

		if (strnicmp(grffile+i, grfname, j)) {
			printf("Warning, this GRD file was generated from %s.GRF.\n", grfname);
			printf("Are you sure you want to apply it to %s?");
			yesno();
		}
	} else {
		strcat(grfname, ".grf");
		grffile = grfname;
	}

	grf = fopen(grffile, "rb");
	if (!grf) die("Can't open %s: %s\n", grffile, strerror(errno));

	tmp = fopen(tempfile, "wb");
	if (!tmp) die("Can't open %s: %s\n", tempfile, strerror(errno));

	printf("Writing temporary file %s\n", tempfile);
  }

  block = malloc(BLOCKSIZE);
  if (!block) die("Out of memory.\n");

  curno = 0;
  for (i=0; i<numsprites; i++) {
	fread(&spriteno, 2, 1, grd);

	if (onlyshow) {
		if (spriteno != lastto+1) {
			if (lastfrom != lastto) printf("-%d", lastto);
			if (lastfrom >= 0) printf(", ");
			lastfrom = spriteno;
			printf("%d", lastfrom);
		}
		lastto = spriteno;

		skipsprite(grd);
	} else {
		printf("\rSprite%5d  Done:%3d%%  \r",
			spriteno, (int) (100L*i/(numsprites-1)));

		while (++curno <= spriteno)
			copysprite(grf, tmp);

		skipsprite(grf);
		copysprite(grd, tmp);
	}
  }

  if (onlyshow) {
	if (lastfrom < 0)
		printf("No sprites.");
	else if (lastfrom != lastto)
		printf("-%d", lastto);
	printf("\n");
  } else {
	// write the EOF and dummy checksum
	magic = 0;
	fwrite(&magic, 2, 1, tmp);
	fwrite(&magic, 4, 1, tmp);
  }

  if (tmp) fclose(tmp);
  if (grd) fclose(grd);
  if (grf) fclose(grf);

  if (!onlyshow) {
	printf("\nDone\n");

	// rename grf to bak if bak doesn't exist
	strcpy(block, grffile);
	c = strrchr(block, '.');
	if (!c) c = block + strlen(block);
	strcpy(c, ".bak");

	tmp = fopen(block, "rb");

	if (!tmp && (errno == ENOENT)) {
		// .bak doesn't exist, rename .grf to .bak
		printf("Renaming %s to %s\n", grffile, block);
		if (rename(grffile, block)) {
			printf("Error while renaming: %s\n", strerror(errno));
			printf("Shall I delete it instead?");
			yesno();
			errno = EEXIST;		// go delete it
		} else {
			errno = ENOENT;		// don't try to delete it
		}
	}
	if (tmp || (errno != ENOENT)) {
		printf("Deleting %s\n", grffile);
		if (remove(grffile))
			die("Error while deleting: %s\n", strerror(errno));
	}

	// rename tmp to grf
	printf("Renaming %s to %s\n", tempfile, grffile);
	if (rename(tempfile, grffile))
		die("Error while renaming: %s\n", strerror(errno));

	printf("All done!\n");
  }

  free(block);
  free(grfname);

  return 1;
}

int main(int argc, char **argv)
{
  char opt;

  printf("GRFMerge - Copyright (C) 2002 by Josef Drexler\n");

#ifdef WIN32
//  debugint();
#endif

  checkisselfextr(argv[0]);

  while (1) {
	opt = getopt(argc, argv, "hly");

	if (opt == (char) EOF)
		break;

	switch (opt) {
		case 'l':
			onlyshow = 1;
			break;
		case 'y':
			alwaysyes = 1;
			break;
		default:
			usage();
	}
  }
  if (!issfx && (optind < argc) )	// no first arg if we're self-extracting
	grdfile = argv[optind++];

  if (optind < argc)
	grffile = argv[optind++];

  if (!grdfile) {
	// no GRD file specified; see if we're a self-extracting executable
	if (issfx)
		grdfile = argv[0];
	else {
		printf("No GRD file specified!\n");
		exit(2);
	}
  }

  return domerge();
}